"""Summarise the cue messages in Tristan data."""

import argparse

import numpy as np

from ..data import (
    cue_id_key,
    cue_keys,
    cue_time_key,
    cues,
    latrd_data,
    reserved,
    seconds,
)
from . import data_files, input_parser, version_parser

parser = argparse.ArgumentParser(
    description=__doc__, parents=[version_parser, input_parser]
)


def main(args=None):
    """Print a human-readable summary of the cue messages in a LATRD data set."""
    args = parser.parse_args(args)

    raw_files, _ = data_files(args.data_dir, args.stem)

    with latrd_data(raw_files, keys=cue_keys) as data:
        relevant = (data[cue_id_key] > 0) & (data[cue_id_key] != reserved)
        cue_ids = data[cue_id_key][relevant].compute()
        cue_times = data[cue_time_key][relevant].compute()

    unique_cues = np.sort(np.unique(cue_ids))

    print("\nSummary of cue messages:")

    for cue in unique_cues:
        cue_description = cues.get(cue, f"Unknown (0x{cue:04x})")
        cues_sel = cue_ids == cue
        cue_times_sel = cue_times[cues_sel]
        deduplicated = np.sort(np.unique(cue_times_sel))

        if deduplicated.size > 1:
            time_diffs = np.diff(deduplicated)
            min_diff = time_diffs.min()
            max_diff = time_diffs.max()
            avg_diff = time_diffs.mean()

            print(
                f"""
{cue_description}:
Found {cue_times_sel.size} instances.
Found {deduplicated.size} de-duplicated instances with
\tsmallest time difference: {min_diff} cycles ({seconds(min_diff):~.3g}),
\tlargest time difference: {max_diff} cycles ({seconds(max_diff):~.3g}),
\tmean time difference: {avg_diff:.2f} cycles ({seconds(avg_diff):~.3g})."""
            )
        elif cue_times_sel.size > 1:
            n = cue_times_sel.size
            print(
                f"\n{cue_description}:  Found {n} instances,\n"
                f"\tall with the same timestamp."
            )
        else:
            print(f"\n{cue_description}:  Found 1 instance.")



"""Tools for creating and manipulating an HDF5 VDS for Tristan data."""

from contextlib import ExitStack
from itertools import chain, compress, cycle, zip_longest
from pathlib import Path
from typing import Dict, Iterable, List

import h5py
import numpy as np

from .data import cue_keys, event_keys, ts_key_regex

Sources = Dict[str, Iterable[h5py.VirtualSource]]
TimeSliceInfo = List[slice], np.ndarray, List[slice], List[slice]
VirtualSourceInfo = Sources, Sources, List[int], Dict[str, type]
Layouts = Dict[str, h5py.VirtualLayout]


def time_slice_info(meta_file: h5py.File) -> TimeSliceInfo:
    """
    Assemble information about the event data time slices from the metadata file.

    Args:
        meta_file:  Metadata ('_meta.h5') file.  Assumes metadata version 1.

    Returns:
        - List of slice objects used to select each time slice from the virtual source
          objects in order to populate the virtual layouts.  Length and order
          correspond to 'events_per_ts'.
        - List of the number of events in each time slice, in the order that the time
          slices will appear in the VDS.  Length is the number of time slices recorded.
    """
    fp_per_module = meta_file["fp_per_module"][()]

    ts_keys = sorted(filter(ts_key_regex.match, meta_file.keys()))
    ts_data = [meta_file[ts_key] for ts_key in ts_keys]

    time_slices = []
    num_events_per_ts = []
    # Loop through the modules, acting on the time slice metadata for each in turn.
    for num_files, ts_counts in zip(fp_per_module, ts_data):
        ts_counts = ts_counts[()]
        # Reshape the time slice metadata for a single module into a rectangular array
        # with shape (number of time slices per file, number of files), so as to be
        # able to generate file-specific slices.
        num_ts_per_fp = -(-ts_counts.size // num_files)
        ts_counts.resize(num_ts_per_fp * num_files)
        # Keep a separate record of each module's array of event counts per time slice.
        num_events_per_ts.append(ts_counts)
        ts_counts = ts_counts.reshape(num_ts_per_fp, num_files)
        # Generate the cumulative count of events per time slice for each file.
        ts_per_module = np.pad(np.cumsum(ts_counts, axis=0), ((1, 0), (0, 0)))
        # Turn these counts into slices to select from a virtual source for each file.
        time_slices.append(
            map(slice, ts_per_module[:-1].flatten(), ts_per_module[1:].flatten())
        )

    # Assemble all the source slices into a single list, ordered first
    # chronologically, then by module number.  Where modules have recorded different
    # numbers of time slices, zip_longest will pad with None.
    time_slices = list(chain.from_iterable(zip_longest(*time_slices)))

    # Resize each module's array of event counts per time slice so that their sizes
    # match.  This is achieved by zero-padding to match the None-padding of the list
    # of time slices by zip_longest.
    max_size = max(data.size for data in num_events_per_ts)
    num_events_per_ts = np.column_stack(
        [np.pad(data, (0, max_size - data.size)) for data in num_events_per_ts]
    ).flatten()

    return time_slices, num_events_per_ts


def virtual_sources(files: List[Path], meta_file: h5py.File) -> VirtualSourceInfo:
    """
    Create HDF5 virtual sources and collate ancillary information from raw data files.

    Args:
        files:      Lexicographically sorted list of raw file paths.
        meta_file:  Tristan detector metadata file object.

    Returns:
        - Dictionary of event data set names and iterators of corresponding HDF5 virtual
          sources.  The iterator of sources for each data set is based on
          itertools.cycle and so repeats indefinitely in the order in which successive
          event slices should be selected to build the virtual data set.
        - Dictionary of cue data set names and lists of corresponding HDF5 virtual
          sources.  The lists of sources have length and order as per the list of input
          files.
        - List of the number of cues in each data file after zero-padding has been
          stripped.  Length and order as per the list of input files.
        - Dictionary of data set names and corresponding data types.
    """
    event_sources = {key: [] for key in event_keys}
    cue_sources = {key: [] for key in cue_keys}
    num_cues_per_file = []

    with ExitStack() as stack:
        raw_files = [stack.enter_context(h5py.File(path, "r")) for path in files]

        dtypes = {key: raw_files[0][key].dtype for key in event_keys + cue_keys}

        for raw_file in raw_files:
            # The cues are padded with zeroes.  Find the first so we can slice them off.
            num_cues_per_file.append(np.argmax(raw_file["cue_id"][()] == 0))
            for key in event_keys:
                event_sources[key].append(h5py.VirtualSource(raw_file[key]))
            for key in cue_keys:
                cue_sources[key].append(h5py.VirtualSource(raw_file[key]))

    # Make a list of slices with which to divide the lexicographically sorted list of
    # file paths into sub-lists, each slice corresponding to a different detector
    # module.  Ordered by module number.  Length is equal to the number of modules in
    # the detector.
    file_slices = np.pad(np.cumsum(meta_file["fp_per_module"]), (1, 0))
    file_slices = list(map(slice, file_slices[:-1], file_slices[1:]))

    # Construct a carousel to select time slices in the order in which they should
    # appear in the virtual layout.
    for key, sources in event_sources.items():
        carousel = zip(*(cycle(sources[file_slice]) for file_slice in file_slices))
        event_sources[key] = chain.from_iterable(carousel)

    return event_sources, cue_sources, num_cues_per_file, dtypes


def virtual_layouts(num_events: int, num_cues: int, dtypes: Dict[str, type]) -> Layouts:
    """Create a dictionary of data set names and corresponding HDF5 virtual layouts."""
    layouts = {}
    for key in event_keys:
        layouts[key] = h5py.VirtualLayout(shape=(num_events,), dtype=dtypes[key])
    for key in cue_keys:
        layouts[key] = h5py.VirtualLayout(shape=(num_cues,), dtype=dtypes[key])
    return layouts


def virtual_data_set(
    raw_files: List[Path],
    meta_file: h5py.File,
    time_slices: List[slice],
    events_per_ts: Iterable[int],
) -> Layouts:
    """
    Define a virtual data set in the form of virtual layouts linked to virtual sources.

    The time slices of events will appear in the VDS ordered chronologically, then by
    module number to resolve ties.  Cue data will simply be aggregated in
    lexicographical order of the raw files from which they are sourced.

    Args:
        raw_files:      Lexicographically sorted list of raw file paths.
        meta_file:      Tristan detector metadata file object.
        time_slices:    List of slice objects used to select each time slice from the
                        virtual source objects in order to populate the virtual
                        layouts.  Length and order correspond to 'events_per_ts'.
        events_per_ts:  List of the number of events in each time slice, in the order
                        that the time slices will appear in the VDS.  Length is the
                        number of time slices recorded.

    Returns:
        A dictionary of raw data set names and corresponding HDF5 virtual layouts.
    """
    event_sources, cue_sources, cues_per_file, dtypes = virtual_sources(
        raw_files, meta_file
    )

    # Generate the slices used to assign the cue data to their virtual layouts.
    cue_slices = np.pad(np.cumsum(cues_per_file), (1, 0))
    num_cues = cue_slices[-1]
    cue_slices = list(map(slice, cue_slices[:-1], cue_slices[1:]))

    # Generate the slices used to assign the event data to their virtual layouts.
    event_slices = np.pad(np.cumsum(events_per_ts), (1, 0))
    num_events = event_slices[-1]
    event_slices = list(map(slice, event_slices[:-1], event_slices[1:]))

    layouts = virtual_layouts(num_events, num_cues, dtypes)

    # Map virtual source slices to virtual layout slices.
    for key, sources in event_sources.items():
        layout = layouts[key]
        # HDF5 VDS doesn't like empty slices, so only assign a source time slice to a
        # layout slice if the corresponding number of events is non-zero. Trying to
        # assign empty slices to empty slices somehow results in a corrupted VDS.
        mapping = compress(zip(event_slices, sources, time_slices), events_per_ts)
        for event_slice, source, time_slice in mapping:
            layout[event_slice] = source[time_slice]
    for key, sources in cue_sources.items():
        layout = layouts[key]
        for cue_slice, source, count in zip(cue_slices, sources, cues_per_file):
            layout[cue_slice] = source[:count]

    return layouts



"""Tools for binning events to images."""


from operator import mul
from typing import Dict, Tuple

import numpy as np
from dask import array as da
from dask.distributed import progress
from numpy.typing import ArrayLike

from .data import (
    event_keys,
    event_location_key,
    event_time_key,
    first_cue_time,
    pixel_index,
    shutter_close,
    shutter_open,
)

Data = Dict[str, da.Array]


def find_start_end(data: Data, distributed: bool = False) -> (int, int):
    """
    Find the shutter open and shutter close timestamps.

    Args:
        data:  A LATRD data dictionary (a dictionary with data set names as keys
               and Dask arrays as values).  Must contain one entry for cue id
               messages and one for cue timestamps.  The two arrays are assumed
               to have the same length.
        distributed:  Whether the computation uses the Dask distributed scheduler,
                      in which case a progress bar will be displayed.

    Returns:
        The shutter open and shutter close timestamps, in clock cycles.
    """
    start_time = first_cue_time(data, shutter_open)
    end_time = first_cue_time(data, shutter_close)

    # If we are using the distributed scheduler (for multiple images), show progress.
    if distributed:
        print("Finding detector shutter open and close times.")
        progress(start_time.persist(), end_time.persist())
    start_time, end_time = da.compute(start_time, end_time)
    print()

    return start_time, end_time


def valid_events(data: Data, start: int, end: int) -> Data:
    """
    Return those events that have a timestamp in the specified range.

    Args:
        data:   An LATRD data dictionary, containing an 'event_time_offset' item
                and optional 'event_id' and 'event_energy' items.
        start:  The start time of the accepted range, in clock cycles.
        end:    The end time of the accepted range, in clock cycles.

    Returns:
        A dictionary containing only the valid events.
    """
    valid = (start <= data[event_time_key]) & (data[event_time_key] < end)

    for key in event_keys:
        value = data.get(key)
        if value is not None:
            data[key] = value[valid].compute_chunk_sizes()

    return data


def make_images(data: Data, image_size: Tuple[int, int], bins: ArrayLike) -> da.Array:
    """
    Bin LATRD events data into images of event counts.

    Given a collection of events data, a known image shape and an array of the
    desired time bin edges, make an image for each time bin, representing the number
    of events recorded at each pixel.

    Args:
        data:        A LATRD data dictionary (a dictionary with data set names as keys
                     and Dask arrays as values).  Must contain one entry for event
                     location messages and one for event timestamps.  The two arrays are
                     assumed to have the same length.
        image_size:  The (y, x), i.e. (slow, fast) dimensions (number of pixels) of
                     the image.
        bins:        The time bin edges of the images (in clock cycles, to match the
                     event timestamps).

    Returns:
        A dask array representing the calculations required to obtain the
        resulting image.
    """
    # We need to ensure that the chunk layout of the event location array matches
    # that of the event time array, so that we can perform matching blockwise iterations
    event_locations = data[event_location_key].rechunk(data[event_time_key].chunks)
    event_locations = pixel_index(event_locations, image_size)

    num_images = len(bins) - 1

    if num_images > 1:
        # We cannot perform a single bincount of the entire data set because that
        # would require allocating enough memory for the entire image stack.

        # Find the index of the image to which each event belongs.
        image_indices = da.digitize(data[event_time_key], bins) - 1

        (images_in_block,) = da.compute(map(np.unique, image_indices.blocks))

        # Construct a stack of images using dask.array.bincount.
        images = []
        for i in range(num_images):
            # When searching for events with a given image index, we already know we
            # can exclude some blocks and thereby save some computation time.
            contains_index = [i in indices for indices in images_in_block]

            image_events = event_locations.blocks[contains_index][
                image_indices.blocks[contains_index] == i
            ]
            images.append(da.bincount(image_events, minlength=mul(*image_size)))

        images = da.stack(images)
    else:
        images = da.bincount(event_locations, minlength=mul(*image_size))

    return images.astype(np.uint32).reshape(num_images, *image_size)



"""Tools for extracting data on cues and events from Tristan data files."""

import re
from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Dict, Iterable, Optional, Tuple, Union

import h5py
from dask import array as da
from dask import config
from numpy.typing import ArrayLike
from pint import Quantity

from . import clock_frequency

# Type alias for collections of raw file paths.
RawFiles = Iterable[Union[str, Path]]

# Regex for the names of data sets, in the time slice metadata file, representing the
# distribution of time slices across raw data files for each module.
ts_key_regex = re.compile(r"ts_qty_module\d{2}")


# Translations of the basic cue_id messages.
padding = 0
sync = 0x800
shutter_open = 0x840
shutter_close = 0x880
fem_falling = 0x8C1
fem_rising = 0x8E1
ttl_falling = 0x8C9
ttl_rising = 0x8E9
lvds_falling = 0x8CA
lvds_rising = 0x8EA
tzero_falling = 0x8CB
tzero_rising = 0x8EB
sync_falling = 0x8CC
sync_rising = 0x8EC
reserved = 0xF00
cues = {
    padding: "Padding",
    sync: "Extended time stamp, global synchronisation",
    shutter_open: "Shutter open time stamp, global",
    shutter_close: "Shutter close time stamp, global",
    fem_falling: "FEM trigger, falling edge",
    fem_rising: "FEM trigger, rising edge",
    ttl_falling: "Clock trigger TTL input, falling edge",
    ttl_rising: "Clock trigger TTL input, rising edge",
    lvds_falling: "Clock trigger LVDS input, falling edge",
    lvds_rising: "Clock trigger LVDS input, rising edge",
    tzero_falling: "Clock trigger TZERO input, falling edge",
    tzero_rising: "Clock trigger TZERO input, rising edge",
    sync_falling: "Clock trigger SYNC input, falling edge",
    sync_rising: "Clock trigger SYNC input, rising edge",
    0xBC6: "Error: messages out of sync",
    0xBCA: "Error: messages out of sync",
    reserved: "Reserved",
    **{
        basic + n: f"{name} time stamp, sensor module {n}"
        for basic, name in (
            (sync, "Extended"),
            (shutter_open, "Shutter open"),
            (shutter_close, "Shutter close"),
        )
        for n in range(1, 64)
    },
}

# Keys of cues and events data in the HDF5 file structure.
cue_id_key = "cue_id"
cue_time_key = "cue_timestamp_zero"
event_location_key = "event_id"
event_time_key = "event_time_offset"
event_energy_key = "event_energy"

cue_keys = cue_id_key, cue_time_key
event_keys = event_location_key, event_time_key, event_energy_key

nx_size_key = "entry/instrument/detector/module/data_size"


def aggregate_chunks(
    existing_chunks: Iterable[int], item_size: int, subdivision: int = 1
):
    target_size_bytes = int(Quantity(config.get("array.chunk-size")).m_as("bytes"))

    # The optimal number of data per Dask chunk.
    target_size = target_size_bytes // item_size

    # Try to aggregate the input data into the fewest possible Dask chunks.
    new_chunks = []
    for chunk in existing_chunks:
        # If this input data set will fit into the current chunk, add it.
        if new_chunks and new_chunks[-1] + chunk <= target_size:
            new_chunks[-1] += chunk
        # If the current chunk is full (or the chunks list is empty), add this
        # data set to the next chunk.
        elif chunk <= target_size:
            new_chunks.append(chunk)
        # If this data set is larger than the max Dask chunk size, split it
        # along the HDF5 data set chunk boundaries and put the pieces in
        # separate Dask chunks.
        else:
            n_whole_chunks, remainder = divmod(chunk, target_size)
            dask_chunk_size = target_size // subdivision * subdivision
            new_chunks += [dask_chunk_size] * n_whole_chunks + [remainder]

    return new_chunks


@contextmanager
def latrd_data(
    raw_file_paths: RawFiles, keys: Iterable[str] = cue_keys + event_keys
) -> Dict[str, da.Array]:
    """
    A context manager to read LATRD data sets from multiple files.

    The yielded dictionary has an entry for each of the specified LATRD data keys.
    Each key must be a valid LATRD data key and the corresponding value is a Dask
    array containing the corresponding LATRD data from all the raw data files,
    rechunked into blocks approximately the size of the default Dask array chunk
    size, but with chunk boundaries aligned with HDF5 data set chunk boundaries.

    Args:
        raw_file_paths:  The paths of the raw LATRD data files.
        keys:  The set of LATRD data keys to be read.

    Yields:
        A dictionary of LATRD data keys and arrays of the corresponding data.
    """
    with ExitStack() as stack:
        files = [
            stack.enter_context(h5py.File(path, "r", swmr=True))
            for path in raw_file_paths
        ]

        data = {}
        for key in keys:
            data_sets = [f[key] for f in files]

            sizes = (d.size for d in data_sets)
            hdf5_chunk_size = (data_sets[0].chunks or (data_sets[0].size,))[0]
            chunks = aggregate_chunks(
                sizes, data_sets[0].dtype.itemsize, hdf5_chunk_size
            )

            data[key] = da.concatenate(data_sets).rechunk(chunks)

        yield data


def first_cue_time(data: Dict[str, da.Array], message: int) -> Optional[da.Array]:
    """
    Find the timestamp of the first instance of a cue message in a Tristan data set.

    Args:
        data:     A LATRD data dictionary (a dictionary with data set names as keys
                  and Dask arrays as values).  Must contain one entry for cue id
                  messages and one for cue timestamps.  The two arrays are assumed
                  to have the same length.
        message:  The message code, as defined in the Tristan standard.

    Returns:
        The timestamp, measured in clock cycles from the global synchronisation signal.
        If the message doesn't exist in the data set, this returns None.
    """
    index = da.argmax(data[cue_id_key] == message)
    if index or data[cue_id_key][0] == message:
        return data[cue_time_key][index]


def cue_times(data: Dict[str, da.Array], message: int) -> da.Array:
    """
    Find the timestamps of all instances of a cue message in a Tristan data set.

    The found timestamps are de-duplicated.

    Args:
        data:     A LATRD data dictionary (a dictionary with data set names as keys
                  and Dask arrays as values).  Must contain one entry for cue id
                  messages and one for cue timestamps.  The two arrays are assumed
                  to have the same length.
        message:  The message code, as defined in the Tristan standard.

    Returns:
        The timestamps, measured in clock cycles from the global synchronisation
        signal, de-duplicated.
    """
    index = da.flatnonzero(data[cue_id_key] == message)
    return da.unique(data[cue_time_key][index])


def seconds(timestamp: int, reference: int = 0) -> Quantity:
    """
    Convert a Tristan timestamp to seconds, measured from a given reference timestamp.

    The time between the provided timestamp and a reference timestamp, both provided
    as a number of clock cycles from the same time origin, is converted to units of
    seconds.  By default, the reference timestamp is zero clock cycles, the beginning
    of the detector epoch.

    Args:
        timestamp:  A timestamp in number of clock cycles, to be converted to seconds.
        reference:  A reference time stamp in clock cycles.

    Returns:
        The difference between the two timestamps in seconds.
    """
    return ((timestamp - reference) / clock_frequency).to_base_units().to_compact()


def pixel_index(location: ArrayLike, image_size: Tuple[int, int]) -> ArrayLike:
    """
    Extract pixel coordinate information from an event location (event_id) message.

    Translate a Tristan event location message to the index of the corresponding
    pixel in the flattened image array (i.e. numbered from zero, in row-major order).

    The pixel coordinates of an event on a Tristan detector are encoded in a 32-bit
    integer location message (the event_id) with 26 bits of useful information.
    Extract the y coordinate (the 13 least significant bits) and the x coordinate
    (the 13 next least significant bits).  Find the corresponding pixel index in the
    flattened image array by multiplying the y value by the size of the array in x,
    and adding the x value.

    This function calls the Python built-in divmod and so can be broadcast over NumPy
    and Dask arrays.

    Args:
        location:    Event location message (an integer).
        image_size:  Shape of the image array in (y, x), i.e. (slow, fast).

    Returns:
        Index in the flattened image array of the pixel where the event occurred.
    """
    x, y = divmod(location, 0x2000)
    # The following is equivalent to, but a little simpler than,
    # return da.ravel_multi_index((y, x), image_size)
    return x + y * image_size[1]



"""General utilities for the command-line tools."""

import argparse
import glob
import re
import sys
from itertools import filterfalse
from pathlib import Path
from typing import List, Optional, SupportsFloat, SupportsInt, Tuple, Union

import h5py
import numpy as np
import pint.errors

from .. import __version__, ureg

__all__ = (
    "triggers",
    "check_output_file",
    "check_multiple_output_files",
    "data_files",
    "version_parser",
    "input_parser",
    "image_output_parser",
    "trigger_parser",
    "exposure_parser",
    "interval_parser",
)

from ..data import (
    fem_falling,
    fem_rising,
    lvds_falling,
    lvds_rising,
    ttl_falling,
    ttl_rising,
)

Quantity, Unit = ureg.Quantity, ureg.Unit

# Regex for Tristan data file name stems.
meta_file_name_regex = re.compile(r"(.*)_(?:meta|\d+)")


triggers = {
    "TTL-rising": ttl_rising,
    "TTL-falling": ttl_falling,
    "LVDS-rising": lvds_rising,
    "LVDS-falling": lvds_falling,
    "FEM-rising": fem_rising,
    "FEM-falling": fem_falling,
}


def check_output_file(
    out_file: Optional[Union[Path, str]] = None,
    stem: Optional[str] = None,
    suffix: str = "output",
    force: bool = False,
) -> Optional[Path]:
    """
    Find and check the output file path.

    Find the output file path, given either a specified output file name or a file
    name stem and suffix.  Exit if any output file(s) already exist, unless instructed
    to overwrite.

    Args:
        out_file:  A suggested output file path.
        stem:  A file name stem to use if out_file is not provided.
        suffix:  A suffix to append to stem when constructing the output file name.
        force:  Choose to overwrite any existing files.

    Returns:
        The output file path.

    Raises:
        SystemExit:  if the output file exists and force is not true.
    """
    if out_file or stem:
        out_file = Path(out_file or f"{stem}_{suffix}.h5").expanduser().resolve()

        if not force and out_file.exists():
            sys.exit(
                f"This output file already exists:\n\t{out_file}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )

        return out_file


def check_multiple_output_files(
    quantity: int,
    out_file: Optional[Union[Path, str]] = None,
    stem: Optional[str] = None,
    suffix: str = "output",
    force: bool = False,
) -> Optional[Tuple[List[Path], Path]]:
    """
    Find and check file paths for output of multiple image sequences.

    Find the output file paths, given either a specified output file name pattern or
    a file name stem and suffix.  Exit if the output file already exists,
    unless instructed to overwrite.

    Args:
        quantity:  THe number of output files to be generated.
        out_file:  A suggested output file path.
        stem:  A file name stem to use if out_file is not provided.
        suffix:  A suffix to append to stem when constructing the output file name.
        force:  Choose to overwrite any existing file of the same name.

    Returns:
        The output file paths.

    Raises:
        SystemExit:  if any output file exists and force is not true.
    """
    if out_file or stem:
        n_dig = len(str(quantity))
        out_file_pattern = (
            Path(out_file or f"{stem}_{suffix}.h5").expanduser().resolve()
        )
        out_files = [
            out_file_pattern.parent / (out_file_pattern.stem + f"_{i + 1:0{n_dig}d}.h5")
            for i in range(quantity)
        ]

        exists = "\n\t".join(map(str, filter(Path.exists, out_files)))
        if not force and exists:
            sys.exit(
                f"The following output files already exist:\n\t{exists}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )

        return out_files, out_file_pattern


def data_files(data_dir: Path, stem: str, n_dig: int = 6) -> (List[Path], Path):
    """
    Extract information about the files containing raw cues and events data.

    Args:
        data_dir: Directory containing the raw data and time slice metadata HDF5 files.
        stem:     Input file name, stripped of '_meta.h5', '_000001.h5', etc..
        n_dig:    Number of digits in the raw file number, e.g. six in '_000001.h5'.

    Returns:
        - Lexicographically sorted list of raw file paths.
        - File path of the time slice metadata file.
    """
    meta_file = data_dir / f"{stem}_meta.h5"
    if not meta_file.exists():
        sys.exit(f"Could not find the expected detector metadata file:\n\t{meta_file}")

    with h5py.File(meta_file, "r") as f:
        n_files = np.sum(f.get("fp_per_module", default=()))

    if n_files:
        raw_files = [data_dir / f"{stem}_{n + 1:0{n_dig}d}.h5" for n in range(n_files)]
        missing = list(filterfalse(Path.exists, raw_files))
        if missing:
            missing = "\n\t".join(map(str, missing))
            sys.exit(f"The following expected data files are missing:\n\t{missing}")
    else:
        print(
            "The detector metadata hold no information about the number of "
            "expected raw data files.  Falling back on finding the data dynamically."
        )
        search_path = str(data_dir / f"{stem}_{n_dig * '[0-9]'}.h5")
        raw_files = [Path(path_str) for path_str in sorted(glob.glob(search_path))]

    return raw_files, meta_file


# A simple version parser.  Print the version and exit.
version_parser = argparse.ArgumentParser(add_help=False)
version_parser.add_argument(
    "-V",
    "--version",
    action="version",
    version="%(prog)s:  Tristan tools {version}".format(version=__version__),
)


class _InputFileAction(argparse.Action):
    """
    From an input file name argument, find the directory and file name stem.

    Set them as attributes of the argument parser instance.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        data_dir, file_name_stem = self.find_input_file_name(values)
        setattr(namespace, "data_dir", data_dir)
        setattr(namespace, "stem", file_name_stem)

    @staticmethod
    def find_input_file_name(in_file: Union[Path, str]) -> (Path, str):
        """
        Resolve the input file name into a directory and a file name stem.

        The file name stem is the file name stem stripped of the last _(meta|\\d+).

        Args:
            in_file:  The input file path.

        Returns:
            The data directory path and the file name stem.
        """
        in_file = Path(in_file).expanduser().resolve()

        if in_file.is_dir():
            data_dir = in_file

            try:
                (file_name,) = data_dir.glob("*_meta.h5")
            except ValueError:
                sys.exit(
                    "Could not find a single unique '<filename>_meta.h5' file in the "
                    "specified directory.\n"
                    "Please specify the desired input file name instead."
                )
            file_name_stem = meta_file_name_regex.fullmatch(file_name.stem)[1]
        else:
            data_dir = in_file.parent

            # Get the segment 'name_stem' from 'name_stem_meta.h5' or
            # 'name_stem_000001.h5'.
            file_name_stem = meta_file_name_regex.fullmatch(in_file.stem)
            if file_name_stem:
                file_name_stem = file_name_stem[1]
            else:
                sys.exit(
                    "Input file name did not have the expected format "
                    "'<name>_meta.h5':\n"
                    f"\t{in_file}"
                )

        return data_dir, file_name_stem


# A parser with the attributes data_dir and stem,
# representing the input Tristan data files.
input_parser = argparse.ArgumentParser(add_help=False)
input_parser.add_argument(
    "input_file",
    help="Tristan metadata ('_meta.h5') or raw data ('_000001.h5', etc.) file.  "
    "This file must be in the same directory as the HDF5 files containing all the "
    "corresponding raw events data.",
    metavar="input-file",
    action=_InputFileAction,
)


def image_size(size: str) -> (int, int):
    """
    Unpack an image size tuple from a comma-separated string of integers.

    Args:
        size:  A string of comma-separated values, representing the size in pixels of
               the image, in the order (x, y).

    Returns:
        The image size tuple in the order (y, x), for compatibility with row-major data.
    """
    x_size, y_size = map(int, size.strip("()'\"").split(","))
    if x_size < 0 or y_size < 0:
        raise ValueError("Image dimensions must not be negative.")
    if not (x_size and y_size):
        raise ValueError("At least one image dimension must be positive.")
    return y_size, x_size


# A parser to specify
#   - a custom output file path,
#   - whether to allow an existing file to be over-written by the output file,
#   - and what size the output images should have.
image_output_parser = argparse.ArgumentParser(add_help=False)
image_output_parser.add_argument(
    "-o",
    "--output-file",
    help="File name or location for output image file, defaults to the working "
    "directory.  If only a directory location is given, the pattern of the raw data "
    "files will be used.  For multiple-sequence output, a sequence number will be "
    "appended to your choice of output file name.",
)
image_output_parser.add_argument(
    "-f",
    "--force",
    help="Force the output image file to overwrite any existing file with the same "
    "name.",
    action="store_true",
)
image_output_parser.add_argument(
    "-s",
    "--image-size",
    help="Dimensions of the detector in pixels, separated by a comma, as 'x,y', i.e. "
    "'fast,slow'.",
    type=image_size,
)


# A parser to specify a trigger signal type.
trigger_parser = argparse.ArgumentParser(add_help=False)
trigger_parser.add_argument(
    "-t",
    "--trigger-type",
    help="The type of trigger signal used as the pump pulse marker.",
    choices=triggers.keys(),
    required=True,
)


def units_of_time(quantity: Union[Quantity, SupportsFloat, str]) -> Quantity:
    """
    Ensure a quantity of time, has compatible units, defaulting to seconds.

    Args:
        quantity:  Any object that can be interpreted as a Pint quantity, with or
                   without units.

    Returns:
        The initial quantity, with units of seconds applied if it was previously
        dimensionless.

    Raises:
        pint.errors.DimensionalityError:  The specified quantity is not dimensionless
                                          and does not have dimension [time].
    """
    # Catch any UndefinedUnitError (which is a subclass of AttributeError) and
    # re-raise it as a ValueError so that argparse knows that this is a bad argument.
    try:
        quantity = Quantity(quantity)
    except pint.errors.UndefinedUnitError as e:
        raise ValueError(e)

    if quantity <= 0:
        raise ValueError("Time quantity must be positive.")

    quantity = quantity * Unit("s") if quantity.dimensionless else quantity
    if quantity.check("[time]"):
        return quantity
    else:
        raise pint.errors.DimensionalityError(
            quantity,
            "a quantity of",
            quantity.dimensionality,
            pint.Unit("s").dimensionality,
        )


def positive_int(value: SupportsInt) -> int:
    """
    Check that an integer value is positive.

    Args:
        value:  A value that can be cast to an integer.

    Returns:
        'value' as an integer, if it is positive.

    Raises:
        ValueError:  If 'value' does not cast to a positive integer.
    """
    int_value = int(value)
    if not int_value > 0:
        raise ValueError(f"The value {value} does not cast to a positive integer.")
    return int_value


# A parser that determines the desired image exposure time, either from an explicit
# specification, or implicitly from a specified number of images.
exposure_parser = argparse.ArgumentParser(add_help=False)
group = exposure_parser.add_mutually_exclusive_group(required=True)
group.add_argument(
    "-e",
    "--exposure-time",
    help="Duration of each image.  This will be used to calculate the number of "
    "images.  Specify a value with units like '--exposure-time .5ms', '-e 500µs' or "
    "'-e 500us'.  Unspecified units default to seconds.",
    type=units_of_time,
)
group.add_argument("-n", "--num-images", help="Number of images.", type=positive_int)


# A parser for subdividing a regular comb of pump signals into quantised pump-probe
# delay intervals.
interval_parser = argparse.ArgumentParser(add_help=False)
group = interval_parser.add_mutually_exclusive_group(required=True)
group.add_argument(
    "-i",
    "--interval",
    help="Duration of each pump-probe delay interval.  This will be used to calculate "
    "the number of image sequences.  Specify a value with units like "
    "'--exposure-time .5ms', '-e 500µs' or '-e 500us'.  Unspecified units default to "
    "seconds.",
    type=units_of_time,
)
group.add_argument(
    "-x", "--num-sequences", help="Number of image sequences.", type=positive_int
)



"""
Create an HDF5 virtual data set (VDS) file to aggregate raw Tristan events data.

By default, this file will be saved in the same directory as the raw data and
detector metadata files, retaining the same naming convention.  So if the metadata
file is named 'my_data_1_meta.h5', then the new VDS file will be named
'my_data_1_vds.h5'.
"""

import argparse

import h5py

from ..vds import time_slice_info, virtual_data_set
from . import check_output_file, data_files, input_parser, version_parser

parser = argparse.ArgumentParser(
    description=__doc__, parents=[version_parser, input_parser]
)
parser.add_argument(
    "-o",
    "--output-file",
    help="File name for output VDS file.  "
    "By default, the pattern of the input file will be used, with '_meta.h5' "
    "replaced with '_vds.h5'.",
)
parser.add_argument(
    "-f",
    "--force",
    help="Force the output file to over-write any existing file with the same name.",
    action="store_true",
)


def main(args=None):
    """Utility for making an HDF5 VDS from raw Tristan data."""
    args = parser.parse_args(args)
    output_file = check_output_file(args.output_file, suffix="vds", force=args.force)

    raw_files, meta_file = data_files(args.data_dir, args.stem)
    with h5py.File(meta_file, "r") as f:
        ts_info = time_slice_info(f)
        layouts = virtual_data_set(raw_files, f, *ts_info)

    with h5py.File(output_file, "w" if args.force else "x") as f:
        for layout in layouts.items():
            f.create_virtual_dataset(*layout)

    print(f"Virtual data set file written to\n\t{output_file}")



# coding: utf-8

"""
Utilities for processing data from the Large Area Time-Resolved Detector

This module provides tools to interpret NeXus-like data in HDF5 format from the
experimental Timepix-based event-mode detector, codenamed Tristan, at Diamond Light
Source.
"""

__author__ = "Diamond Light Source - Scientific Software"
__email__ = "scientificsoftware@diamond.ac.uk"
__version__ = "0.1.12"
__version_tuple__ = tuple(int(x) for x in __version__.split("."))

import pint

ureg = pint.UnitRegistry()

clock_frequency = ureg.Quantity(6.4e8, "Hz").to_compact()



#!/usr/bin/env python

"""Apply a flat-field correction to reconstructed Tristan images."""

import argparse
import pathlib
import shutil
import sys
from operator import mul, truediv
from typing import List, Optional

import h5py
import hdf5plugin
import numpy as np
from dask import array as da
from dask.diagnostics import ProgressBar

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument(
    "input_file",
    help="A file of reconstructed Tristan images (.h5 or .nxs), containing an image "
    "data set '/data'.",
    metavar="input-file",
    type=pathlib.Path,
)
parser.add_argument(
    "flat_field_file",
    help="A Tristan flat-field correction (.h5 or .nxs) file, containing the "
    "flat-field measurement in a data set '/image'.",
    metavar="flat-field-file",
    type=pathlib.Path,
)
choices = ["multiply", "divide"]
parser.add_argument(
    "method",
    help="Choose whether to multiply or divide the images by the flat-field "
    "correction data set.  Default is 'divide'.",
    choices=choices,
    default="multiply",
)
parser.add_argument(
    "-o",
    "--output-file",
    help="Path for the output file (.h5 or .nxs).  If not provided, this will default "
    "to the input filename with the added suffix '_flat_field_multiplied' or "
    "'_flat_field_divided'.",
    type=pathlib.Path,
)
parser.add_argument(
    "-f",
    "--force",
    help="Force the output image file to overwrite any existing file with the same "
    "name.",
    action="store_true",
)


file_exists = (
    "Output file already exists:\n\t"
    "{}\n"
    "Use '-f' to override, or specify a different output file path with '-o'."
)


def main(args: Optional[List[str]] = None) -> None:
    """
    Apply the flat field data set to an images data set.

    Args:
        args:  Input command line arguments.  If None, defaults to sys.argv[1:].
    """
    args = parser.parse_args(args)

    method = dict(zip(choices, ["multiplied", "divided"]))[args.method]
    output_file = args.output_file or args.input_file.stem + f"_flat_field_{method}.h5"
    output_file = pathlib.Path(output_file).with_suffix(".h5")
    write_mode = "w" if args.force else "x"

    try:
        with h5py.File(args.input_file.with_suffix(".h5")) as f, h5py.File(
            args.flat_field_file
        ) as g, h5py.File(output_file, write_mode) as h, ProgressBar():
            images = da.from_array(f["data"])
            flat_field = g["image"]
            # Multiply or divide the images by the flat-field correction.
            func = dict(zip(choices, [mul, truediv]))[args.method]
            images = func(images, np.where(flat_field, flat_field, 1))
            images = images.astype(np.uint32)
            h.require_dataset(
                "data",
                shape=images.shape,
                dtype=images.dtype,
                chunks=images.chunksize,
                **hdf5plugin.Bitshuffle(),
            )
            images.store(h["data"])
    except FileExistsError:
        sys.exit(file_exists.format(output_file))

    output_nexus = output_file.with_suffix(".nxs")
    if output_nexus.exists() and not args.force:
        sys.exit(file_exists.format(output_nexus))
    else:
        try:
            shutil.copy(args.input_file.with_suffix(".nxs"), output_nexus)
            with h5py.File(output_nexus, "r+") as f:
                del f["entry/data/data"]
                f["entry/data/data"] = h5py.ExternalLink(str(output_file), "data")
                f["entry/instrument/detector/flatfield_applied"][()] = "TRUE"
        except FileNotFoundError:
            sys.exit("Could not find input NeXus file to copy.")



"""Aggregate the events from a LATRD Tristan data collection into one or more images."""

import argparse
import sys
from pathlib import Path
from typing import Iterable, Tuple, Union

import dask
import h5py
import numpy as np
import pint
import zarr
from dask import array as da
from dask import delayed
from dask.diagnostics import ProgressBar
from dask.distributed import Client, progress
from dask.system import CPU_COUNT
from hdf5plugin import Bitshuffle
from nexgen.nxs_copy import CopyTristanNexus

from .. import clock_frequency
from ..binning import find_start_end, make_images, valid_events
from ..data import (
    aggregate_chunks,
    cue_keys,
    cue_times,
    cues,
    event_location_key,
    event_time_key,
    first_cue_time,
    latrd_data,
    seconds,
)
from . import (
    check_multiple_output_files,
    check_output_file,
    data_files,
    exposure_parser,
    image_output_parser,
    input_parser,
    interval_parser,
    trigger_parser,
    triggers,
    version_parser,
)


def determine_image_size(nexus_file: Path) -> Tuple[int, int]:
    """Find the image size from metadata."""
    recommend = f"Please specify the detector dimensions in (x, y) with '--image-size'."
    try:
        with h5py.File(nexus_file, "r") as f:
            return f["entry/instrument/detector/module/data_size"][()]
    except (FileNotFoundError, OSError):
        sys.exit(f"Cannot find NeXus file:\n\t{nexus_file}\n{recommend}")
    except KeyError:
        sys.exit(f"Input NeXus file does not contain image size metadata.\n{recommend}")


def exposure(
    start: int, end: int, exposure_time: pint.Quantity = None, num_images: int = None
) -> (pint.Quantity, int, int):
    if exposure_time:
        exposure_time = exposure_time.to_base_units().to_compact()
        exposure_cycles = (exposure_time * clock_frequency).to_base_units().magnitude
        num_images = int((end - start) // exposure_cycles)
    else:
        # Because they are expected to be mutually exclusive, if there is no
        # exposure_time, there must be a num_images.
        exposure_cycles = int((end - start) / num_images)
        exposure_time = seconds(exposure_cycles)

    return exposure_time, exposure_cycles, num_images


def single_image_cli(args):
    """Utility for making a single image from event-mode data."""
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(
        args.output_file, args.stem, "single_image", args.force
    )
    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.single_image_nexus(
                output_file, input_nexus, write_mode=write_mode
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    with latrd_data(raw_files, keys=cue_keys) as data:
        start, end = find_start_end(data)

    print("Binning events into a single image.")
    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        image = make_images(valid_events(data, start, end), image_size, (start, end))

        with ProgressBar(), h5py.File(output_file, write_mode) as f:
            data_set = f.require_dataset(
                "data",
                shape=image.shape,
                dtype=image.dtype,
                chunks=image.chunksize,
                **Bitshuffle(),
            )
            image.store(data_set)

    print(f"Image written to\n\t{output_nexus or output_file}")


def save_multiple_images(
    array: da.Array, output_file: Path, write_mode: str = "x"
) -> None:
    """
    Calculate and store a Dask array in an HDF5 file without exceeding available memory.

    Use the Dask distributed scheduler to compute a Dask array and store the
    resulting values to a data set 'data' in the root group of an HDF5 file.  The
    distributed scheduler is capable of managing worker memory better than the
    default scheduler.  In the latter case, the workers can sometimes demand more
    than the available amount of memory.  Using the distributed scheduler avoids this
    problem.

    The distributed scheduler cannot write directly to HDF5 files because h5py.File
    objects are not serialisable.  To work around this issue, the data are first
    stored to a Zarr DirectoryStore, then copied to the final HDF5 file and the Zarr
    store deleted.

    Multithreading is used, as the calculation is assumed to be I/O bound.

    Args:
        array:  A Dask array to be calculated and stored.
        output_file:  Path to the output HDF5 file.
        write_mode:  HDF5 file opening mode.  See :class:`h5py.File`.
    """
    # Set a more generous connection timeout than the default 30s.
    with dask.config.set(
        {
            "distributed.comm.timeouts.connect": "60s",
            "distributed.comm.timeouts.tcp": "60s",
            "distributed.deploy.lost-worker-timeout": "60s",
            "distributed.scheduler.idle-timeout": "600s",
            "distributed.scheduler.locks.lease-timeout": "60s",
        }
    ):
        intermediate = str(output_file.with_suffix(".zarr"))

        # Overwrite any pre-existing Zarr storage.  Don't compute immediately but
        # return the Array object so we can compute it with a progress bar.
        method = {"overwrite": True, "compute": False, "return_stored": True}
        # Prepare to save the calculated images to the intermediate Zarr store.
        array = array.to_zarr(intermediate, component="data", **method)
        # Compute the Array and store the values, using a progress bar.
        progress(array.persist())

    print("\nTransferring the images to the output file.")
    store = zarr.DirectoryStore(intermediate)
    with h5py.File(output_file, write_mode) as f:
        zarr.copy_all(zarr.open(store), f, **Bitshuffle())

    # Delete the Zarr store.
    store.clear()


def save_multiple_image_sequences(
    array: da.Array,
    intermediate_store: Union[Path, str],
    output_files: Iterable[Path],
    write_mode: str = "x",
) -> None:
    intermediate_store = Path(intermediate_store).with_suffix(".zarr")

    # Set a more generous connection timeout than the default 30s.
    with dask.config.set(
        {
            "distributed.comm.timeouts.connect": "60s",
            "distributed.comm.timeouts.tcp": "60s",
            "distributed.deploy.lost-worker-timeout": "60s",
            "distributed.scheduler.idle-timeout": "600s",
            "distributed.scheduler.locks.lease-timeout": "60s",
        }
    ):
        # Overwrite any pre-existing Zarr storage.  Don't compute immediately but
        # return the Array object so we can compute it with a progress bar.
        method = {"overwrite": True, "compute": False, "return_stored": True}
        # Prepare to save the calculated images to the intermediate Zarr store.
        array = [
            sub_array.to_zarr(intermediate_store, component=f"{i:d}/data", **method)
            for i, sub_array in enumerate(array)
        ]
        # Compute the Array and store the values, using a progress bar.
        progress([sub_array.persist() for sub_array in array])
        print()

    print("Transferring the images to the output files.")
    store = zarr.DirectoryStore(str(intermediate_store))
    arrays = zarr.open(store)

    @delayed
    def sequence_to_disk(i, output_file):
        with h5py.File(output_file, write_mode) as f:
            return zarr.copy_all(arrays[i], f, **Bitshuffle())

    transfer = [sequence_to_disk(i, o).persist() for i, o in enumerate(output_files)]
    progress(transfer)
    da.compute(transfer)

    print()

    # Delete the Zarr store.
    store.clear()


def multiple_images_cli(args):
    """
    Utility for making multiple images from event-mode data.

    The time between the start and end of the data collection is subdivided into a
    number of exposures of equal duration, providing a chronological stack of images.
    """
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(args.output_file, args.stem, "images", args.force)

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if not input_nexus.exists():
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    with latrd_data(raw_files, keys=cue_keys) as data:
        start, end = find_start_end(data, distributed=True)
        exposure_time, exposure_cycles, num_images = exposure(
            start, end, args.exposure_time, args.num_images
        )

        if args.align_trigger:
            trigger_type = triggers.get(args.align_trigger)
            print(
                f"Image start and end times will be chosen such that the first "
                f"'{cues[trigger_type]}' is aligned with an image boundary."
            )
            # Note we are assuming that the first trigger time is after shutter open.
            trigger_time = first_cue_time(data, trigger_type)
            if trigger_time is None:
                sys.exit(f"Could not find a '{cues[trigger_type]}' signal.")
            trigger_time = trigger_time.compute().astype(int)
            bins_pre = np.arange(
                trigger_time - exposure_cycles, start, -exposure_cycles, dtype=np.uint64
            )[::-1]
            bins_post = np.arange(trigger_time, end, exposure_cycles, dtype=np.uint64)
            bins = np.concatenate((bins_pre, bins_post))
        else:
            bins = np.linspace(start, end, num_images + 1, dtype=np.uint64)

    print(
        f"Binning events into {num_images} images with an exposure time of "
        f"{exposure_time:~.3g}."
    )

    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        images = make_images(valid_events(data, start, end), image_size, bins)
        save_multiple_images(images, output_file, write_mode)

    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.multiple_images_nexus(
                output_file,
                input_nexus,
                nbins=num_images,
                write_mode=write_mode,
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None

    print(f"Images written to\n\t{output_nexus or output_file}")


def pump_probe_cli(args):
    """
    Utility for making multiple images from a pump-probe data collection.

    The time between one pump trigger signal and the next is subdivided into a number
    of exposures of equal duration.  Data from all such pump-to-pump intervals is
    aggregated, providing a single stack of images that captures the evolution of the
    response of the measurement to a pump signal.
    """
    write_mode = "w" if args.force else "x"
    output_file = check_output_file(args.output_file, args.stem, "images", args.force)

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if input_nexus.exists():
        try:
            # Write output NeXus file if we have an input NeXus file.
            output_nexus = CopyTristanNexus.pump_probe_nexus(
                output_file, input_nexus, write_mode=write_mode
            )
        except FileExistsError:
            sys.exit(
                f"This output file already exists:\n\t"
                f"{output_file.with_suffix('.nxs')}\n"
                "Use '-f' to override, "
                "or specify a different output file path with '-o'."
            )
    else:
        output_nexus = None
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    trigger_type = triggers.get(args.trigger_type)

    print("Finding trigger signal times.")
    with latrd_data(raw_files, keys=cue_keys) as data:
        trigger_times = cue_times(data, trigger_type)
        progress(trigger_times.persist())
        trigger_times = trigger_times.compute().astype(int)

    print()  # Dask distributed progress bar does not end with a newline, so insert one.

    trigger_times = np.sort(trigger_times)

    if not trigger_times.any():
        sys.exit(f"Could not find a '{cues[trigger_type]}' signal.")

    end = np.diff(trigger_times).min()
    exposure_time, _, num_images = exposure(0, end, args.exposure_time, args.num_images)
    bins = np.linspace(0, end, num_images + 1, dtype=np.uint64)

    print(
        f"Binning events into {num_images} images with an exposure time of "
        f"{exposure_time:~.3g} according to the time elapsed since the most recent "
        f"'{cues[trigger_type]}' signal."
    )

    trigger_times = da.from_array(trigger_times)
    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        # Measure the event time as time elapsed since the most recent trigger signal.
        data[event_time_key] = data[event_time_key].astype(np.int64)
        data[event_time_key] -= trigger_times[
            da.digitize(data[event_time_key], trigger_times) - 1
        ]

        images = make_images(valid_events(data, 0, end), image_size, bins)
        save_multiple_images(images, output_file, write_mode)

    print(f"Images written to\n\t{output_nexus or output_file}")


def multiple_sequences_cli(args):
    """
    Utility for making multiple image sequences from a pump-probe data collection.

    The time between one pump trigger signal and the next is subdivided into a number
    of intervals of equal duration, quantising the time elapsed since the most recent
    trigger pulse.  Events are labelled according to which interval they fall in and,
    for each interval in turn, all the events so labelled are aggregated, providing a
    stack of image sequences that captures the evolution of the response of the
    measurement to a pump signal.
    """
    write_mode = "w" if args.force else "x"

    input_nexus = args.data_dir / f"{args.stem}.nxs"
    if not input_nexus.exists():
        print(
            "Could not find a NeXus file containing experiment metadata.\n"
            "Resorting to writing raw image data without accompanying metadata."
        )

    image_size = args.image_size or determine_image_size(input_nexus)

    raw_files, _ = data_files(args.data_dir, args.stem)

    trigger_type = triggers.get(args.trigger_type)

    print("Finding trigger signal times.")
    with latrd_data(raw_files, keys=cue_keys) as data:
        trigger_times = cue_times(data, trigger_type)
        progress(trigger_times.persist())
        trigger_times = trigger_times.compute().astype(int)

    print()  # Dask distributed progress bar does not end with a newline, so insert one.

    trigger_times = np.sort(trigger_times)

    if not trigger_times.any():
        sys.exit(f"Could not find a '{cues[trigger_type]}' signal.")

    intervals_end = np.diff(trigger_times).min()
    interval_time, _, num_intervals = exposure(
        0, intervals_end, args.interval, args.num_sequences
    )
    intervals = np.linspace(0, intervals_end, num_intervals + 1, dtype=np.uint64)

    output_files, out_file_pattern = check_multiple_output_files(
        num_intervals, args.output_file, args.stem, "images", args.force
    )

    with latrd_data(raw_files, keys=cue_keys) as data:
        start, end = find_start_end(data, distributed=True)

    exposure_time, exposure_cycles, num_images = exposure(
        start, end, args.exposure_time, args.num_images
    )
    bins = np.linspace(start, end, num_images + 1, dtype=np.uint64)

    print(
        f"Using '{cues[trigger_type]}' as the pump signal,\n"
        f"binning events into {num_intervals} sequences, corresponding to "
        f"successive pump-probe delay intervals of {interval_time:~.3g}.\n"
        f"Each sequence consists of {num_images} images with an exposure time of "
        f"{exposure_time:~.3g}."
    )

    out_file_stem = out_file_pattern.stem

    n_dig = len(str(num_images))
    out_file_pattern = out_file_pattern.parent / f"{out_file_stem}_{'#' * n_dig}.h5"

    if input_nexus.exists():
        # Write output NeXus files if we have an input NeXus file.
        output_nexus_pattern = out_file_pattern.with_suffix(".nxs")
        for output_file in output_files:
            try:
                CopyTristanNexus.multiple_images_nexus(
                    output_file,
                    input_nexus,
                    nbins=num_images,
                    write_mode=write_mode,
                )
            except FileExistsError:
                sys.exit(
                    f"One or more output files already exist, matching the pattern:\n\t"
                    f"{output_nexus_pattern}\n"
                    "Use '-f' to override, "
                    "or specify a different output file path with '-o'."
                )
    else:
        output_nexus_pattern = None

    trigger_times = da.from_array(trigger_times)
    with latrd_data(raw_files, keys=(event_location_key, event_time_key)) as data:
        data = valid_events(data, start, end)

        # Find the time elapsed since the most recent trigger signal.
        pump_probe_time = data[event_time_key].astype(np.int64)
        pump_probe_time -= trigger_times[
            da.digitize(pump_probe_time, trigger_times) - 1
        ]
        sequence = da.digitize(pump_probe_time, intervals) - 1

        image_sequence_stack = []
        for i in range(num_intervals):
            interval_selection = sequence == i

            event_times = data[event_time_key][interval_selection]
            event_locs = data[event_location_key][interval_selection]
            interval = {
                event_time_key: event_times.compute_chunk_sizes(),
                event_location_key: event_locs.compute_chunk_sizes(),
            }

            size = max(data[event_time_key].itemsize, data[event_location_key].itemsize)
            chunks = aggregate_chunks(*interval[event_time_key].chunks, size)
            interval[event_time_key] = interval[event_time_key].rechunk(chunks)

            image_sequence_stack.append(make_images(interval, image_size, bins))

        save_multiple_image_sequences(
            da.stack(image_sequence_stack), out_file_stem, output_files, write_mode
        )

    print(f"Images written to\n\t{output_nexus_pattern or out_file_pattern}")


parser = argparse.ArgumentParser(description=__doc__, parents=[version_parser])
subparsers = parser.add_subparsers(
    help="Choose the manner in which to create images.",
    required=True,
    dest="sub-command",
)

parser_single = subparsers.add_parser(
    "single",
    aliases=["1"],
    description=(
        "Aggregate all the events from a LATRD Tristan data collection "
        "into a single image."
    ),
    parents=[version_parser, input_parser, image_output_parser],
)
parser_single.set_defaults(func=single_image_cli)

parser_multiple = subparsers.add_parser(
    "multiple",
    aliases=["multi"],
    description=(
        "Bin the events from a LATRD Tristan data collection into multiple images."
    ),
    parents=[version_parser, input_parser, image_output_parser, exposure_parser],
)
parser_multiple.add_argument(
    "-a",
    "--align-trigger",
    help="Align the start and end time of images such that the first trigger signal of "
    "the chosen type is matched up with an image start time.  Useful for examining "
    "effects in the data before and after a single trigger pulse.",
    choices=triggers.keys(),
)
parser_multiple.set_defaults(func=multiple_images_cli)

parser_pump_probe = subparsers.add_parser(
    "pump-probe",
    aliases=["pp"],
    description="Bin events into images representing different pump-probe delays.\n\n"
    "With LATRD data from a pump-probe experiment, where the pump signal has a fairly "
    "constant repeat rate, bin events into a stack of images spanning the range of "
    "pump-probe delay times, from shortest to longest.",
    parents=[
        version_parser,
        input_parser,
        image_output_parser,
        trigger_parser,
        exposure_parser,
    ],
)
parser_pump_probe.set_defaults(func=pump_probe_cli)

parser_multiple_sequences = subparsers.add_parser(
    "sequences",
    aliases=["sweeps"],
    description="Bin events into several sequences of images, each corresponding to "
    "a different pump-probe delay time interval.\n\n"
    "With LATRD data from a pump-probe experiment, where the pump signal has a fairly "
    "constant repeat rate, the recorded events are separated into groups corresponding "
    "to the time elapsed since the most recent pump trigger signal.  Each group is "
    "binned into a sequence of chronological images.  Each sequence is saved to a "
    "separate output file, numbered from the shortest pump-probe delay to the longest.",
    parents=[
        version_parser,
        input_parser,
        image_output_parser,
        trigger_parser,
        exposure_parser,
        interval_parser,
    ],
)
parser_multiple_sequences.set_defaults(func=multiple_sequences_cli)


def main(args=None):
    """Perform the image binning with a user-specified sub-command."""
    args = parser.parse_args(args)
    if args.func == single_image_cli:
        #  Use the default scheduler.
        args.func(args)
    else:  # Multi-image binning requires the Dask distributed scheduler.
        # Use threads, rather than processes.
        with Client(processes=False, threads_per_worker=int(0.9 * CPU_COUNT) or 1):
            args.func(args)
